/**********************************
  Aggregate Rate Controller (ARC)
  Author: Jae Chung
  Email:  goos@cs.wpi.edu
  Date:   Sep. 2003
***********************************/

#include <math.h>
#include <string.h>
#include <sys/types.h>
#include "template.h"
#include "random.h"
#include "flags.h"
#include "delay.h"
#include "arcq.h"

/*
 * ARC OTcl linkage class
 */
static class ARCClass : public TclClass {
public:
    ARCClass() : TclClass("Queue/ARC") {}
    TclObject* create(int, const char*const*) {
	return (new ARCQueue);
    }
} class_arc;

/*
 * When update_timer_ expires call ARCQueue:update_stats()
 */
void ARCUpdateTimer::expire(Event*)
{
    t_->update_stats();
}

/*
 * ARCQueue construtor
 */
ARCQueue::ARCQueue() : link_(NULL), de_drop_(NULL), tchan_(0),
		       bcount_(0), update_timer_(this)
{
    bind_bool("setbit_", &arcp_.setbit);            /* mark instead of drop */
    bind_bool("queue_in_bytes_", &arcp_.qib);       /* use q in bytes?  */
    bind_bool("bytes_", &arcp_.bytes);              /* use bytes option? */
    bind("mean_pktsize_", &arcp_.mean_pktsize);     /* avg pkt size (bytes) */
    bind("update_interval_",&arcp_.d);              /* update interval */
    bind("alpha_", &arcp_.alpha);                   /* integral control parm */
    bind("target_load_", &arcp_.tload);             /* target load */
    bind("target_q_size_", &arcp_.tqs);             /* target queue size */

    bind("curq_", &curq_);            /* current queue size */
    bind("esr_", &arcv_.esr);         /* estimated service rate */
    bind("notif_p_", &arcv_.notif_p); /* congestion notification probability */
    
    /* if the update interval is negative or zero, set it to 1.0 sec */
    if(arcp_.d <= 0) {
	printf("Warning: update_interval_ <= 0. Set to 1.0 sec\n");
	arcp_.d = 1.0;
    }
    
    /* q_ and pq_ will be used for packet queue */
    q_ = new PacketQueue();
    pq_ = q_;
    
    /* reset the queue */
    reset();
    
#ifdef debug_arc
    print_arcp();
    print_arcv();
#endif
    
}


/*
 * ARCQueue reset
 */
void ARCQueue::reset()
{
    /* Set service rate (in bytes per second) */
    if (link_) {
	arcv_.sr = (double)link_->bandwidth() / 8.;
    } else {
	arcv_.sr = 1250000; /* default = 10Mbps */
    }
    
    arcv_.esr = 0.0;
    arcv_.notif_p = 0.0;
    
    arcv_.b_svcd  = 0;
    
    curq_   = 0;
    bcount_ = 0;
    
    Queue::reset();
    
    update_timer_.resched(arcp_.d);
}


/*
 * Update Congestion Notification Probability
 */
void ARCQueue::update_stats()
{
    /* Queue minus target queue (in bytes) */
    int b_que = arcp_.qib ? bcount_ : q_->length() * arcp_.mean_pktsize;
    b_que -= arcp_.tqs * arcp_.mean_pktsize;
    if(b_que < 0) b_que = 0;

    /* Trace: Estimated Service Rate (in bps) */
    arcv_.esr = (double)(arcv_.b_svcd * 8) / arcp_.d;

    /* Calculate notif_p */
    double notif_p = arcv_.notif_p;
    notif_p += arcp_.alpha 
               * (arcv_.b_svcd - arcp_.tload * (arcp_.d * arcv_.sr - b_que))
               / arcp_.mean_pktsize;

    if(notif_p > 1) notif_p = 1;
    if(notif_p < 0) notif_p = 0;
    arcv_.notif_p = notif_p;

    /* Reset bytes serviced */
    arcv_.b_svcd = 0;
    
    /* Reschedule next update */
    update_timer_.resched(arcp_.d);
}


/*
 * Return the next packet in the queue for transmission.
 */
Packet* ARCQueue::deque()
{
    Packet *p;
    p = q_->deque();
    if (p != 0) 
	bcount_ -= hdr_cmn::access(p)->size();
    
    return (p);
}


/*
 * Generate events with given probability
 */
int ARCQueue::notify_early(double notif_p)
{
    double u = Random::uniform();
    if (u <= notif_p)
	return(1);
    else
	return(0);
}

/*
 * Enqueue packets with early congestion notification
 */
void ARCQueue::enque(Packet* pkt)
{
    hdr_cmn* cmh  = hdr_cmn::access(pkt);
    int qlim = arcp_.qib ? (qlim_ * arcp_.mean_pktsize) : qlim_;
    int qlen = arcp_.qib ? bcount_ : q_->length(); 
    curq_ = qlen;	/* help tracing queue, if enabled */


    /* check queue overflow */
    if (qlen >= qlim) {
	drop(pkt);
    }
    else {
	/* if "bytes" option, adjust p based on the packet size */
	double p = arcv_.notif_p;
	if (arcp_.bytes) {
	    p = p * cmh->size() / arcp_.mean_pktsize;
	    if (p > 1) p = 1;
	}
	
	/* early congestion notification */
	if(notify_early(p)) {
	    hdr_flags* hf = hdr_flags::access(pkt);
	    if (arcp_.setbit && hf->ect()) {
		/* mark the ECN bit */
		hf->ce() = 1;
	    } else {
		/* implicit drop notification */
		if (de_drop_ != NULL)
		    de_drop_->recv(pkt);
		else
		    drop(pkt);
		return;
	    }   
	}

	/* enqueue the packet and record bytes */
	q_->enque(pkt);
	bcount_ += cmh->size();
	arcv_.b_svcd += cmh->size();
    }

    return;
}


int ARCQueue::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
    if (argc == 2) {
	if (strcmp(argv[1], "reset") == 0) {
	    reset();
	    return (TCL_OK);
	}
	if (strcmp(argv[1], "early-drop-target") == 0) {
	    if (de_drop_ != NULL)
		tcl.resultf("%s", de_drop_->name());
	    return (TCL_OK);
	}
    } else if (argc == 3) {
	/* attach a file for variable tracing */
	if (strcmp(argv[1], "attach") == 0) {
	    int mode;
	    const char* id = argv[2];
	    tchan_ = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
	    if (tchan_ == 0) {
		tcl.resultf("ARC: trace: can't attach %s for writing", id);
		return (TCL_ERROR);
	    }
	    return (TCL_OK);
	}
	/* tell ARC about link stats */
	if (strcmp(argv[1], "link") == 0) {
	    LinkDelay* del = (LinkDelay*)TclObject::lookup(argv[2]);
	    
	    if (del == 0) {
		tcl.resultf("ARC: no LinkDelay object %s", argv[2]);
		return(TCL_ERROR);
	    }
	    /* calc service rates (bps) */
	    link_ = del;
	    arcv_.sr = (double)link_->bandwidth() / 8.;
	    return (TCL_OK);
	}
	if (strcmp(argv[1], "early-drop-target") == 0) {
	    NsObject* p = (NsObject*)TclObject::lookup(argv[2]);
	    if (p == 0) {
		tcl.resultf("no object %s", argv[2]);
		return (TCL_ERROR);
	    }
	    de_drop_ = p;
	    return (TCL_OK);
	}
	if (!strcmp(argv[1], "packetqueue-attach")) {
	    delete q_;
	    if (!(q_ = (PacketQueue*) TclObject::lookup(argv[2])))
		return (TCL_ERROR);
	    else {
		pq_ = q_;
		return (TCL_OK);
	    }
	}
    }
    return (Queue::command(argc, argv));
}

/*
 * Routine called by TracedVar facility when variables change values.
 * Currently used to trace values of avg queue size, drop probability,
 * and the instantaneous queue size seen by arriving packets.
 * Note that the tracing of each var must be enabled in tcl to work.
 */

void ARCQueue::trace(TracedVar* v)
{
    char wrk[500], *p;
    
    if (((p = strstr(v->name(), "esr")) == NULL) &&
	((p = strstr(v->name(), "notif_p")) == NULL) &&
	((p = strstr(v->name(), "curq")) == NULL)) {
	fprintf(stderr, "ARC:unknown trace var %s\n",
		v->name());
	return;
    }
    
    if (tchan_) {
	int n;
	double t = Scheduler::instance().clock();
	// XXX: be compatible with nsv1 RED trace entries
	if (*p == 'c') {
	    sprintf(wrk, "Q %g %d", t, int(*((TracedInt*) v)));
	}
	else if (*p == 'n') {
	    sprintf(wrk, "p %g %g", t, double(*((TracedDouble*) v)));
	}
	else {
	    sprintf(wrk, "%c %g %g", *p, t,
		    double(*((TracedDouble*) v)));
	}
	n = strlen(wrk);
	wrk[n] = '\n'; 
	wrk[n+1] = 0;
	(void)Tcl_Write(tchan_, wrk, n+1);
    }
    return; 
}


/*
 * For debugging help 
 */

void ARCQueue::print_arcp()
{
    printf("ECN:             %d\n", arcp_.setbit);
    printf("Queue In Byte:   %d\n", arcp_.qib); 
    printf("Mean Pkt Size:   %d\n", arcp_.mean_pktsize);
    printf("Update Interval: %lf\n", arcp_.d); 
    printf("=========\n");
}

void ARCQueue::print_arcv()
{
    printf("Current Queue: %d\n", (int)curq_);
    printf("ESR:           %lf\n", (double)arcv_.esr);
    printf("Notif_P        %lf\n", (double)arcv_.notif_p);
    printf("=========\n");
}
